
#include "cubemap.h"
#include <GL/glew.h>
#include <GL/glu.h>
#include "../stb_image.h"
#include "../errors.h"
#include "../glutil.h"
#include <fmt/core.h>

Cubemap::Cubemap(std::vector<std::string> texture_faces) {
    /* faces must be provided in the order:
     *
     * right, left, top, bottom, front, back
     */
    glActiveTexture(GL_TEXTURE11);

    int width, height, nrChannels;

    stbi_set_flip_vertically_on_load(false);

    glGenTextures(1, &id);
    glBindTexture(GL_TEXTURE_CUBE_MAP, id);

    for (unsigned int i = 0; i < texture_faces.size(); i++) {
        unsigned char *data = stbi_load(texture_faces[i].c_str(), 
                &width, &height, &nrChannels, 0);

        if (data) {
            glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X + i, 0, GL_RGB, 
                    width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, data);
            glCheckError();
            stbi_image_free(data);
        } else {
            A_ERROR(fmt::format("Load cubemap face {}\n", texture_faces[i]));
            stbi_image_free(data);
        }
    }
    
    glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);
}



